__author__ = "Tony C. Wu (@verysure), Felix Strieth-Kalthoff (@felix-s-k), Martin Seifrid (@mseifrid)"

import time
from math import ceil
from abc import ABC, abstractmethod
from typing import Union, Optional, Dict


class SyringePump(ABC):
    # TODO: Should SyringePump have a logger?
    """
    Abstract Base Class for handling different kinds of syringe pumps.
    """
    category="Pump"
    ui_fields  = ("com_port", "address", "syringe_volume", "num_valve_port", "init_valve", "out_valve", "home_pos")
    def __init__(self,
                 address: Union[int, str],
                 syringe_volume: float,
                 num_valve_port: int,
                 init_valve: int,
                 out_valve: int,
                 top_speed_ml: float,
                 home_pos: int,
                 top_speed: float,
                 start_speed: float,
                 stop_speed: float,
                 speed_slope: float,
                 wait_time: float,
                 ports: Optional[Dict[str, int]] = None,
                 ):
        """
        take a dict define name of each port, e.g. {'N2': 1, 'water': 2} for ease of use
        Parameters
        ----------
        ports
        """
        if (init_valve > num_valve_port + 1) or (out_valve > num_valve_port + 1):
            raise ValueError("Init_valve or Out valve_num exceed num_valve_port!")
        self.address = address
        self.syringe_volume = syringe_volume
        self.num_valve_port = num_valve_port
        self.init_valve = init_valve
        self.out_valve = out_valve
        self.home_pos = home_pos
        self._top_speed_def = top_speed
        self._top_speed_ml_def = top_speed_ml
        self._start_speed_def = start_speed
        self._stop_speed_def = stop_speed
        self._speed_slope_def = speed_slope
        self.wait_time = wait_time
        self.ports: Optional[Dict[str, int]] = ports if ports else {}

    # Public Methods for Using the Syringe Pump
    def draw(self, volume: float, valve_port: Optional[Union[int, str]] = None, speed: Optional[float] = None) -> None:
        """
        Draws a specific volume through a given valve port.

        Args:
            :param volume: Volume to draw (in mL).
            :param valve_port: Identifier of the valve port to be used.
            :param speed: draw speed (in mL/s)
        """
        if volume == 0:
            return
        assert volume > 0, "Draw volume must be positive"
        current_volume = self.volume
        if (volume + current_volume) > (1.001e3 * self.syringe_volume):
            raise ValueError("Draw volume excess syringe size")
        if speed is not None:
            self.top_speed_ml = speed
        else:
            self.set_default_speeds()
        #     self.top_speed_ml = self.top_speed_ml
        if valve_port is not None:
            self.switch_port(port=valve_port)
            # valve_port: int = self.switch_port(valve_port)
            # self.port = valve_port
        self.volume = volume + current_volume
        # self.top_speed = self.top_speed_ml
        self.top_speed_ml = self._top_speed_ml_def

    def dispense(self, volume: float, valve_port: Optional[Union[int, str]] = None,
                 speed: Optional[float] = None) -> None:
        """
        Dispenses a specific volume through a given valve port.

        Args:
            :param volume: Volume to draw (in mL).
            :param valve_port: Identifier of the valve port to be used.
            :param  speed:  dispense speed (in mL/s)
        """
        if volume == 0:
            return
        assert volume > 0, "Dispense volume must be positive"
        current_volume = self.volume
        if volume > (current_volume + self.syringe_volume):
            # assess if going to dispense too much, the syringe_volume is added to avoid rounding issue
            # it is actually volume in mL / 1000, but as in setting the volume are given in L, it is multiplied back
            raise ValueError('Dispense volume excess amount in syringe')
        if speed is not None:
            self.top_speed_ml = speed
        else:
            self.set_default_speeds()
            # self.top_speed = self.top_speed_ml
        if valve_port is not None:
            self.switch_port(port=valve_port)
            # valve_port: int = self.switch_port(valve_port)
            # self.port = valve_port
        self.volume = current_volume - volume
        # self.top_speed = self.top_speed_ml
        self.top_speed_ml = self._top_speed_ml_def

    def draw_and_dispense(
            self,
            volume: float,
            draw_valve_port: Union[int, str] = None,
            dispense_valve_port: Union[int, str] = None,
            speed: Optional[float] = None,
            wait: float = 0
    ) -> None:
        """
        Draws a specified amount of volume from a specified port, and dispenses it to a specified port.
        Can temporarily change the draw/dispense velocity, if specified.

        Args:
             :param draw_valve_port: Identifier of the valve port to be used for drawing liquid.
             :param dispense_valve_port: Identifier of the valve port to be used for dispensing liquid.
             :param volume: Volume to be drawn and dispensed (in mL).
             :param wait: Waiting time between draw and dispense (in sec).
             :param speed: Draw / dispense velocity to be set temporarily (in mL/s).

        """
        # Multiple aspirations, if the dispense volume is greater than the syringe volume
        dispense_iterations = ceil(volume / (1e3 * self.syringe_volume))
        volume_per_iteration = volume / dispense_iterations
        for i in range(0, dispense_iterations):
            self.draw(volume=volume_per_iteration, valve_port=draw_valve_port, speed=speed)
            time.sleep(wait)
            self.dispense_all(valve_port=dispense_valve_port, speed=speed)
            time.sleep(wait)

    def draw_full(self, **kwargs) -> None:
        current_volume = self.volume
        draw_volume = self.syringe_volume * 1e3 - current_volume
        self.draw(volume=draw_volume, **kwargs)

    def dispense_all(self, **kwargs) -> None:
        current_volume = self.volume
        self.dispense(volume=current_volume, **kwargs)

    def switch_port(self, port: Union[str, int]) -> int:
        """
        If applicable, converts the human identifier for a port (as key in the ports attribute) to the respective
        computer identifier.

        Args:
            port: Human identifier for the port.

        Returns:
            int: Port number, as required for communication with the pump.
        """
        if isinstance(port, str):
            port = self.ports.get(port)
        if port is not None:
            self.port = port
            return port
        else:
            raise ValueError('Wrong ports')

    # Private Methods -> Valve Position
    @abstractmethod
    def port(self, port: int, **kwargs) -> int:
        """
        Abstract method for setting the valve to a specified port.

        Args:
            port: Integer value of the valve position.
        """
        pass

    # Private Methods -> Piston Movement and Position
    @abstractmethod
    def volume(self, volume: float, **kwargs) -> float:
        pass

    @abstractmethod
    def top_speed_ml(self) -> float:
        """
        get the speed of the plunger based on mL/s
        Returns
        -------

        """
        pass

    @abstractmethod
    def set_default_speeds(self):
        """
        set the pump to default speed
        Returns
        -------

        """
        pass

    # def __getattr__(self, attr):
    #     if attr in self.info:
    #         return self.info[attr]
    #     else:
    #         raise AttributeError(f'{attr} does not exist in obj or obj.info')
    #

class PeristalicPump(ABC):
    def __init__(self,
                 address: Union[int, str],
                #  syringe_volume: float,
                #  num_valve_port: int,
                #  init_valve: int,
                #  out_valve: int,
                #  top_speed_ml: float,
                #  home_pos: int,
                #  top_speed: float,
                #  start_speed: float,
                #  stop_speed: float,
                #  speed_slope: float,
                #  wait_time: float,
                #  ports: Optional[Dict[str, int]] = None,
                 ):
        """
        take a dict define name of each port, e.g. {'N2': 1, 'water': 2} for ease of use
        Parameters
        ----------
        ports
        """
        self.address = address
        
    @abstractmethod
    def set_pump(self,
                 speed: float,
                 start: bool,
                 direction_CCW: bool
                 ):
        pass

    @abstractmethod
    def rpm(self)->float:
        pass

    @abstractmethod
    def on(self)-> bool:
        pass

    @abstractmethod
    def direction(self)->bool:
        pass

class ContinuousSyringePump(ABC):
    def __init__(self,
                 com_port:str
                 ):
        pass

    @abstractmethod
    def speed(self)->int:
        pass


from __future__ import annotations
import time
from typing import Dict, Union

from matterlab_serial_device import SerialDevice, open_close
from matterlab_pumps.base_pump import PeristalicPump


class LongerPeristalticPump(PeristalicPump, SerialDevice):
    category="Pump"
    ui_fields  = ("com_port", "address")
    """
    A peristaltic pump controller that communicates over a serial interface.

    Inherits from PeristalicPump for pump logic and SerialDevice for serial comms.

    Attributes:
        _rpm (float): Current pump speed in revolutions per minute.
        _on (bool): Pump running state.
        _direction (bool): Pump direction (True for CCW, False for CW).
    """

    def __init__(
        self,
        com_port: str,
        address: int,
        encoding: str = "utf-8",
        baudrate: int = 9600,
        timeout: float = 1.0,
        parity: str = "even",
        bytesize: int = 8,
        stopbits: int = 1,
        share_port: bool = True,
        **kwargs,
    ) -> None:
        """
        Initialize the pump controller.

        Args:
            com_port (str): Serial port identifier (e.g., 'COM3' or '/dev/ttyUSB0').
            address (int): Device address for Modbus protocol.
            encoding (str, optional): Serial text encoding. Defaults to 'utf-8'.
            baudrate (int, optional): Communication speed. Defaults to 9600.
            timeout (float, optional): Read/write timeout in seconds. Defaults to 1.0.
            parity (str, optional): Parity setting ('even', 'odd', 'none'). Defaults to 'even'.
            bytesize (int, optional): Number of data bits (5, 6, 7, or 8). Defaults to 8.
            stopbits (int, optional): Number of stop bits (1, 1.5, or 2). Defaults to 1.
            **kwargs: Additional keyword args for SerialDevice.
        """
        SerialDevice.__init__(
            self,
            com_port=com_port,
            encoding=encoding,
            baudrate=baudrate,
            parity=parity,
            timeout=timeout,
            bytesize=bytesize,
            stopbits=stopbits,
            share_port=share_port
            **kwargs,
        )
        PeristalicPump.__init__(self, address=address)
        self._rpm: float = 0.0
        self._on: bool = False
        self._direction: bool = False

    def _generate_execute_command(
        self,
        rpm: float,
        on: bool,
        direction: bool,
    ) -> bytes:
        """
        Build the Modbus command to on/stop the pump at the desired speed and direction.

        Args:
            rpm (float): Desired speed in RPM.
            on (bool): True to on the pump, False to stop.
            direction (bool): True for counter-clockwise rotation.

        Returns:
            bytes: Full command packet ready for transmission.
        """
        # Header: [address, function=0x06, register_high=0x57, register_low=0x4A]
        cmd = bytearray([self.address, 0x06, 0x57, 0x4A])
        # Speed is scaled by 10 (1 digit after decimal)
        speed_int = int(rpm * 10)
        cmd.extend([(speed_int >> 8) & 0xFF, speed_int & 0xFF])
        # Flags: on and direction
        cmd.append(int(on))
        cmd.append(int(direction))
        # Append checksum (XOR of all bytes)
        checksum = 0
        for b in cmd:
            checksum ^= b
        cmd.append(checksum)
        # Prefix with 0xE9
        return b"\xE9" + bytes(cmd)

    def _generate_query_command(self) -> bytes:
        """
        Build the Modbus command to query the current pump state.

        Returns:
            bytes: Full query packet ready for transmission.
        """
        cmd = bytearray([self.address, 0x02, 0x52, 0x4A])
        checksum = 0
        for b in cmd:
            checksum ^= b
        cmd.append(checksum)
        return b"\xE9" + bytes(cmd)

    @open_close
    def set_pump(
        self,
        rpm: float | None = None,
        on: bool | None = None,
        direction: bool | None = None,
    ) -> None:
        """
        Update pump settings and send command to device.

        Args:
            rpm (float, optional): Target speed in RPM. If None, previous speed is retained.
            on (bool, optional): True to on, False to stop. If None, previous state is retained.
            direction (bool, optional): True for CCW, False for CW. If None, previous direction is retained.
        """
        rpm = self._rpm if rpm is None else rpm
        on = self._on if on is None else on
        direction = self._direction if direction is None else direction

        packet = self._generate_execute_command(
            rpm=rpm, on=on, direction=direction
        )
        self.query(write_command=packet, return_bytes=True, num_bytes=6)

    @open_close
    def query_pump(self) -> Dict[str, Union[float, bool]]:
        """
        Query the pump's current status and update internal state.

        Returns:
            dict: Dictionary with keys 'rpm', 'on', 'direction'.
        """
        packet = self._generate_query_command()
        response = self.query(write_command=packet, return_bytes=True, num_bytes=10)
        # Small delay to allow device processing
        time.sleep(0.5)
        # Parse speed: two bytes at positions 5 and 6
        speed_raw = (response[5] << 8) | response[6]
        self._rpm = speed_raw / 10.0
        # Parse flags at positions 7 and 8
        self._on = bool(response[7])
        self._direction = bool(response[8])
        return {"rpm": self._rpm, "on": self._on, "direction": self._direction}

    @property
    def rpm(self) -> float:
        """
        Get the latest RPM reading from the pump.
        """
        return self.query_pump()["rpm"]

    @rpm.setter
    def rpm(self, value: float) -> None:
        """
        Set a new pump speed (RPM) without changing other states.
        """
        self.set_pump(rpm=value)

    @property
    def on(self) -> bool:
        """
        Get the pump's running state.
        """
        return self.query_pump()["on"]

    @on.setter
    def on(self, value: bool) -> None:
        """
        on or stop the pump without changing speed or direction.
        """
        self.set_pump(on=value)

    @property
    def direction(self) -> bool:
        """
        Get the pump's current rotation direction (True for CCW).
        """
        return self.query_pump()["direction"]

    @direction.setter
    def direction(self, value: bool) -> None:
        """
        Change the pump's rotation direction without altering speed or runtime state.
        """
        self.set_pump(direction=value)



import pytest
from pathlib import Path
from matterlab_pumps import TecanXCPump
import time

COM_PORT = "COM5"
def test_init():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    assert isinstance(pump, TecanXCPump)
    assert pump.initialization_status == True
    assert pump.syringe_volume == 1e-3
    assert pump.address == "1"

def test_port():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    pump.port = 12
    assert pump.port == 12
    pump.port = 1
    assert pump.port == 1

def test_volume():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    assert pump.volume == 0
    pump.volume = 1.0
    assert pump.volume == 1.0
    pump.volume = 0.0
    assert pump.volume == 0

def test_speed():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    pump.top_speed_ml = 0.5
    assert 0.499 <= pump.top_speed_ml <= 0.501

def test_draw_dispense():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    pump.draw(0.25, 1, speed=0.5)
    assert pump.volume == 1.0
    assert pump.port == 1
    assert 0.499 <= pump.top_speed_ml <= 0.501
    pump.dispense(0.25, 1)
    assert pump.volume == 0.0
    assert pump.port == 1
    assert 0.149 <= pump.top_speed_ml <= 0.151

def test_draw_full_dispense_all():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    pump.port = 1
    pump.draw_full()
    assert pump.volume == 1.0
    pump.dispense_all()
    assert pump.volume == 1.0

def test_draw_and_dispense():
    pump = TecanXCPump(com_port=COM_PORT, address=0, syringe_volume=1e-3, num_valve_port=12)
    pump.draw_and_dispense(volume=1, draw_valve_port=12, dispense_valve_port=1, speed = 0.5, wait = 10)
    assert pump.port == 1
    assert pump.volume == 0




from struct import pack
from matterlab_serial_device import SerialDevice, open_close
from matterlab_pumps.base_pump import ContinuousSyringePump


class ContinuousDualSyringe(ContinuousSyringePump, SerialDevice):
    category="Pump"
    ui_fields  = ("com_port")
    def __init__(self, com_port):
        SerialDevice.__init__(self,
                              com_port=com_port
                              )
        ContinuousSyringePump.__init__(self)
        self._speed = 0

    @property
    def speed(self)->int:
        return self._speed

    @speed.setter
    def speed(self, speed: int):
        self.write(pack(">H", speed))
        


import time
from typing import Dict, Union, Tuple, Optional, List
from matterlab_serial_device import SerialDevice, open_close
from matterlab_pumps.base_pump import SyringePump


class LongerPump(SyringePump, SerialDevice):
    category="Pump"
    ui_fields  = ("com_port", "address", "syringe_volume", "num_valve_port", "init_valve", "out_valve", "home_pos")
    def __init__(self,
                 com_port: str,
                 address: Union[str, int],
                 syringe_volume: float,
                 num_valve_port: int,
                 ports: Optional[Dict[str, int]] = None,
                 init_valve: int = 1,
                 out_valve: int = 12,
                 connect_hardware: bool = True,
                 encoding: str = "utf-8",
                 baudrate: int = 9600,
                 timeout: float = 0.2,
                 bytesize: int = 8,
                 parity: str = "none",
                 stopbits: int = 1,
                 top_speed_ml: float = 0.1,
                 direction_CW: bool = True,
                 fine_pos: bool = False,
                 home_pos: int = 20,
                 backlash_pos: int = 12,
                 top_speed: int = 1000,
                 start_speed: int = 800,
                 stop_speed: int = 800,
                 speed_slope: int =14,
                 wait_time: float = 3600.0,
                 share_port: bool = True,
                 error_msg: List = [
                                    "No Error",
                                    "Initialization Error",
                                    "Invalid Command",
                                    "Invalid Operand",
                                    "Invalid Command Sequence",
                                    "Error Code Not Used",
                                    "EEPROM Failure",
                                    "Device Not Initialized",
                                    "Error Code Not Used",
                                    "Plunger Overload",
                                    "Valve Overload",
                                    "Plunger Move Not Allowed",
                                    "Error Code Not Used",
                                    "Error Code Not Used",
                                    "Error Code Not Used",
                                    "Command Overflow"
                                  ],
                 **kwargs
                 ):
        """
        :param com_port: COM port of the pump connected to, example: "COM1", "/tty/USB0"
        :param address: address of pump, set at the back of the pump, char '0' - 'E' or int 0-9
        :param syringe_volume: syringe volume in unit of L (NOT mL!!!)
        :param num_valve_port: number of valve ports for selection/distribution
        :param ports: dict containing nickname and valve port number mapping
        :param init_valve: valve port num to draw from when initializing, typically air/N2, pure solvent or waste
        :param out_valve: valve port num to dispense to when initializing, typically waste
        :param connect_hardware: if initialize hardware up on connect
        :param encoding: encoding of command, default utf-8
        :param baudrate: baudrate for communication, default 9600
        :param timeout: timeout for communication, default 0.2 s
        :param bytesize: bytesize for communication, default 8
        :param parity: parity for communication, default none
        :param stopbits: stopbits for communication, default 1
        :param top_speed_ml: default speed for transfer, for 1/16 0.03 tubing, 0.1 mL/s
        :param direction_CW: direction of selection valve indexing, default True for clockwise
        :param fine_pos: turn on/off fine positioning, default True
        :param home_pos: home position, default 20
        :param backlash_pos: backlash steps, default 12
        :param top_speed: top speed by steps, default 1000
        :param start_speed: start speed by steps, default 800
        :param stop_speed: stop speed by steps, default 800
        :param speed_slope: speed slop grading, default 14
        :param wait_time: max wait time for a transfer, default 3600 s or 1 h
        :param error_msg: error messages from pump
        :param kwargs:
        """
        SerialDevice.__init__(self,
                              com_port=com_port,
                              encoding=encoding,
                              baudrate=baudrate,
                              timeout=timeout,
                              parity=parity,
                              bytesize=bytesize,
                              stopbits=stopbits,
                              share_port=share_port,
                              **kwargs,
                              )
        SyringePump.__init__(self,
                             address=address,
                             syringe_volume=syringe_volume,
                             num_valve_port=num_valve_port,
                             init_valve=init_valve,
                             out_valve=out_valve,
                             top_speed_ml=top_speed_ml,
                             home_pos=home_pos,
                             top_speed=top_speed,
                             start_speed=start_speed,
                             stop_speed=stop_speed,
                             speed_slope=speed_slope,
                             wait_time=wait_time,
                             ports=ports
                             )
        self.direction_CW = direction_CW
        self.fine_pos = fine_pos
        self.backlash_pos = backlash_pos
        self.error_msg = error_msg
        # Tecan syringe pump use its own way of addressing RS485 device, and is used as self.address in code
        # so initialize the RS485 device with common address format, then update the self.address
        address = self._update_address(address)
        self.address = address

        self.initialization_status = False
        if connect_hardware:
            self.connect()

    def connect(self):
        """
        hardware initialization function
        :return:
        """
        self._set_fine_position()
        self._set_backlash_position()
        self._set_home_position()
        self.initialize_pump()
        self.set_top_speed(top_speed=self._top_speed_def, auto_start_stop_speed=True)
        self.set_speed_slope(speed_slope=self._speed_slope_def)

    @staticmethod
    def _update_address(address: Union[str, int]) -> str:
        """
        update the address setting
        translate   int 0, 1, 2,... 8, 9, 10, ...14
                    str '0', '1', '2',... '8', '9', 'A',... 'E'
            to      chr '1', '2', '3',... '9', ':', ';',... '?'
        Returns
        -------

        """
        if isinstance(address, int):
            return chr(49 + address)
        elif isinstance(address, str):
            return chr(int(('0x' + address), 0) + 49)
        else:
            raise TypeError('Invalid pump address.')

    @open_close
    def _query_pump(self, command: str) -> str:
        """
        send query command to pump
        return sliced return value in str
        Parameters
        ----------
        command

        Returns
        -------

        """
        return self.query(write_command=f"/{self.address}{command}\r",
                          remove_from_start=3, remove_from_end=3, read_delay= 0.05)

    @open_close
    def _execute_pump(self, command: str):
        """
        execute command
        R for 'RUN'
        Parameters
        ----------
        command

        Returns
        -------

        """
        self.write(f"/{self.address}{command}R\r")
        time.sleep(0.05)

    @open_close
    def _busy_report(self):
        """
        check if pump is ready for next command
        set self._ready to True if pump is ready
        set self._ready to False if pump is not ready
        raise IOError if pump report error
        Returns
        -------

        """
        rtn = self.query(write_command=f"/{self.address}Q\r",
                         read_delay=0,
                         read_until=b"\r\n",
                         remove_from_end=2,
                         return_bytes = True,
                         )
        if len(rtn) < 3:  # no meaningful return means communication failure, pump is not ready
            self._ready = False
            return
        if rtn[2] == 0x60:  # 3rd byte is 0x60 (int 96) means pump is ready
            self._ready = True
            return
        else:
            self._ready = False  # pump is not ready
        if rtn[2] != 0x40:  # 3rd byte is 0x40 (int 64) means pump is simply not ready
            error_num = rtn[2] % 16  # not 0x40 means hardware error, raise error message
            raise IOError(self.error_msg[error_num])

    def _ensure_ready(self, wait_time: float = None) -> bool:
        """
        ensure the pump is ready
        Parameters
        ----------
        wait_time: max time to wait until pump is ready

        Returns: True if pump is ready
        -------

        """
        if not wait_time:
            wait_time = self.wait_time
        t0 = time.time()
        while (time.time() - t0) < wait_time:
            self._busy_report()
            if self._ready:
                return True
            else:
                time.sleep(0.1)
        raise IOError(f"Pump busy for {wait_time} seconds and not ready, exiting...")

    def _set_fine_position(self):
        """
        set the fine position of the pump
        if fine position, range (0, 240001), else (0, 3001) NOTE: five folds
        can ONLY be called before initialization!
        Returns
        -------

        """
        self.fine_pos=False
        print("Fine position not available for Longer, set to False ")
        if self.fine_pos:
            self._max_step = 24000
            self.home_pos *= 8
            self.backlash_pos *= 8
            self._execute_pump(command='N1')
        else:
            self._max_step = 3000
            self._execute_pump(command='N0')
        print(f"Fine position is {str(self.fine_pos)}")

    def report_backlash_position(self):
        rtn = int(self._query_pump(command='?12'))
        print(f"Backlash_position is {rtn}.")
        return rtn

    def _update_backlash_position(self, backlash_position):
        self._backlash_position = backlash_position

    def _set_backlash_position(self):
        """
        set the backlash position of the pump
        can ONLY be called before initialization!
        Parameters
        ----------

        Returns
        -------

        """
        self._execute_pump(command=f'K{self.backlash_pos}')
        if self.report_backlash_position() == self.backlash_pos:
            self._update_backlash_position(self.backlash_pos)
        else:
            raise IOError("Set backlash position failed.")

    def report_home_position(self):
        rtn = int(self._query_pump(command='?24'))
        print(f"Home_position is {rtn}.")
        return rtn

    def _update_home_position(self, position):
        self._home_position = position

    def _set_home_position(self):
        """
        set the home position of the pump
        can only be called BEFORE initializing a pump
        Parameters
        ----------

        Returns
        -------

        """
        # TODO: add two-step set home physically
        self._execute_pump(command=f'k{self.home_pos}')
        if self.report_home_position() == self.home_pos:
            self._update_home_position(self.home_pos)
        else:
            raise IOError("Set home position failed.")

    def report_start_speed(self):
        rtn = int(self._query_pump(command='?1'))
        #print(f"Start speed is {rtn}.")
        return rtn

    def _update_start_speed(self, start_speed):
        self._start_speed = start_speed

    def set_start_speed(self, start_speed: int = 800):
        """
        set the start speed of the pump, range (50, 1001)
        Parameters
        ----------
        start_speed

        Returns
        -------

        """
        if start_speed < 50:
            start_speed = 50
        if start_speed > 1000:
            start_speed = 1000
        self._execute_pump(command=f'v{start_speed}')
        if self.report_start_speed() == start_speed:
            self._update_start_speed(start_speed)
        else:
            raise IOError("Set start speed failed!")

    def report_top_speed(self):
        rtn = int(self._query_pump(command='?2'))
        #print(f"Top speed is {rtn}.")
        return rtn

    def _update_top_speed(self, top_speed):
        self._top_speed = top_speed

    def set_top_speed(self, top_speed: int = 1000, auto_start_stop_speed: bool = True):
        """
        set the start speed of the pump, range (50, 6001)
        Notes: top_speed need to set to 2701 at beginning to avoid set start/stop speed failure
        Parameters
        ----------
        top_speed

        Returns
        -------

        """
        self._execute_pump(command=f'V2701')
        if top_speed < 5:
            top_speed = 5
        if top_speed > 5000:
            top_speed = 5000
        if auto_start_stop_speed:
            self.set_start_speed(start_speed=int(top_speed * 0.8))
            self.set_stop_speed(stop_speed=top_speed)
        self._execute_pump(command=f'V{top_speed}')
        if self.report_top_speed() == top_speed:
            self._update_top_speed(top_speed)
        else:
            raise IOError("Set top speed failed!")

    def report_stop_speed(self):
        rtn = int(self._query_pump(command='?3'))
        #print(f"Stop speed is {rtn}.")
        return rtn

    def _update_stop_speed(self, stop_speed):
        self._stop_speed = stop_speed

    def set_stop_speed(self, stop_speed: int = 800):
        """
        set the stop speed of the pump, range (50, 2701)
        Parameters
        ----------
        stop_speed

        Returns
        -------

        """
        if stop_speed < 50:
            stop_speed = 50
        if stop_speed > 2700:
            stop_speed = 2700
        self._execute_pump(command=f'c{stop_speed}')
        if self.report_stop_speed() == stop_speed:
            self._update_stop_speed(stop_speed)
        else:
            raise IOError("Set stop speed failed!")

    def set_speed_slope(self, speed_slope: int = 14):
        """
        set the speed slope, higher accelerate faster
        Parameters
        ----------
        speed_slope

        Returns
        -------

        """
        assert 1 <= speed_slope < 21, "Speed slope out of range"
        self._execute_pump(command=f'L{speed_slope}')

    def set_default_speeds(self):
        self.set_start_speed(int(self._start_speed_def))
        self.set_stop_speed(int(self._stop_speed_def))
        self.set_speed_slope(int(self._speed_slope_def))
        self.set_top_speed(int(self._top_speed_def))

    def report_buffer_status(self):
        self._buffer_status = bool(int(self._query_pump(command='?10')))

    def report_number_initialization(self):
        self._number_initialization = int(self._query_pump(command='?15'))
        print(f"Number_initialization is {self._number_initialization}.")

    def report_plunger_move_counting(self):
        self._plunger_move = int(self._query_pump(command='?16'))
        print(f"Plunger_move is {self._plunger_move}.")

    def report_valve_move_counting(self):
        self._valve_move = int(self._query_pump(command='?17'))
        print(f"Valve_move is {self._valve_move}.")

    def report_firmware_version(self):
        print(f"Firmware version is {self._query_pump(command='23')}")

    def report_dead_volume(self):
        print(f"Dead volume in steps is {self._query_pump(command='24')}")

    def report_pump_configuration(self):
        print("Pump configuration query not available with Longer")
        # print(f"Pump configuration is {self._query_pump(command='76')}")

    def report_plunger_absolute_position(self) -> int:
        rtn = int(self._query_pump(command='?'))
        #print(f"Absolute plunger position is {rtn}.")
        return rtn

    def _update_plunger_absolute_position(self, plunger_absolute_position: int):
        self._plunger_absolute_position = plunger_absolute_position

    def report_plunger_relative_position(self) -> int:
        rtn = int(self._query_pump(command='?4'))
        #print(f"Relative plunger position is {rtn}.")
        return rtn

    def _update_plunger_relative_position(self, plunger_relative_position: int):
        self._plunger_relative_position = plunger_relative_position

    def report_valve_number(self) -> int:
        rtn = int(self._query_pump(command='?6'))
        #print(f"Valve number is {rtn}.")
        return rtn

    def _update_valve_number(self, valve_number: int):
        self._valve_number = valve_number

    def initialize_pump(self, init_valve: int = None, out_valve: int = None):
        """
        initialize the pump
        first check if re-initialize pump that has been initialized
        then set direction and force of initialization
            cmd_head = Z for clockwise initialization, meaning port 1 to port X increase CW,
                    = Y for counter clockwise initialization
            cmd_force = 0 for >= 1 mL syringe
                                = 1 for 100 uL - 1 mL syringe
                                = 2 for <= 100 uL syringe
        initialize using
            init_valve as input, out_valve as output
        communicate with pump after initialization, if ready proceed
            max try 30 times, 1 sec per trial
        set initialization status to True
        Returns
        -------

        """
        if self.initialization_status:
            while True:
                verify = input(
                    'The pump has been initialized, are you sure to re-initialize?\n'
                    'Type Y for initialization, '
                    'N for continue without execution.\n')
                if verify == 'Y':
                    print('Pump re-initialization!')
                    break
                elif verify == 'N':
                    print('Pump not re-initialized!')
                    return None
                else:
                    continue
        if self.direction_CW:
            cmd_head = 'Z'
        else:
            cmd_head = 'Y'
        if 1.0e-3 < self.syringe_volume <= 5.0e-3:
            cmd_force = '0'
        elif 1.0e-4 < self.syringe_volume <= 1.0e-3:
            cmd_force = '1'
        elif 0 < self.syringe_volume <= 1.0e-4:
            cmd_force = '2'
        else:
            raise ValueError("Wrong syringe size, check setting file!")
        if init_valve is None:
            init_valve = self.init_valve
        if out_valve is None:
            out_valve = self.out_valve
        self._execute_pump(command=f'{cmd_head}{cmd_force},{init_valve},{out_valve}')
        for ready_count in range(0, 30):
            if self._ensure_ready(wait_time=30):
                self._update_plunger_absolute_position(self.report_plunger_absolute_position())
                self._update_valve_number(self.report_valve_number())
                self.initialization_status = True
                print(f'Pump {self.address} has been initialized')
                return
            else:
                time.sleep(5)
        raise IOError(f'Pump {self.address} initialization failed!')

    def _initialize_plunger(self):
        """
        initialize the plunger only, use with caution!
            cmd_force = 0 for >= 1 mL syringe
                                = 1 for 100 uL - 1 mL syringe
                                = 2 for <= 100 uL syringe
        Returns
        -------

        """
        input('Caution!\nInitialization of the plunger only, press Enter to confirm...')
        cmd_head = 'W'
        if 1.0e-3 < self.syringe_volume <= 5.0e-3:
            cmd_force = '0'
        elif 1.0e-4 < self.syringe_volume <= 1.0e-3:
            cmd_force = '1'
        elif 0 < self.syringe_volume <= 1.0e-4:
            cmd_force = '2'
        else:
            raise ValueError("Wrong syringe size, check setting file!")
        self._execute_pump(command=f'{cmd_head}{cmd_force}')
        self.initialization_status = False
        print("Plunger initialized only, a full initialization is required for operations.")

    def _initialize_valve(self, init_valve: int = None):
        """
        initialize the valve only, use with caution!
        Returns
        -------

        """
        if self.direction_CW:
            direction = '0'
        else:
            direction = '1'
        if init_valve is None:
            init_valve = self.init_valve
        self._execute_pump(command=f'w{init_valve},{direction}')
        self.initialization_status = False
        print("Plunger initialized only, a full initialization is required for operations.")

    def move_valve(self, valve_num: int,
                   direction_CW: bool = None,
                   confirmation: bool = True,
                   ready_before_execute: int = 600):
        """
        wait max ready_before_execute seconds for pump to be ready
        move valve to valve_num
        if direction_CW is provided, use T for clockwise, F for CCW
            else use settings.direction_CW
        Parameters
        ----------
        valve_num
        direction_CW
        confirmation
        ready_before_execute

        Returns
        -------

        """
        assert self.initialization_status, f'Pump {self.address} has not been initialized.'
        if direction_CW is None:
            direction_CW = self.direction_CW
        if ready_before_execute > 0:
            self._ensure_ready(ready_before_execute)
        if direction_CW:
            self._execute_pump(f'I{valve_num}')
        else:
            self._execute_pump(f'O{valve_num}')
        # print(f'Valve moving to {valve_num}...')
        self._ensure_ready(wait_time=3)
        if confirmation:
            if self.report_valve_number() == valve_num:
                self._update_valve_number(valve_num)
                print(f'Pump {self.address} valve moved to {self._valve_number}.')
            else:
                raise IOError(f'Pump {self.address} move valve failed!')
        else:
            self._update_valve_number(self.report_valve_number())

    def move_plunger(self, plunger_pos: int,
                     confirmation: bool = True,
                     ready_before_execute: int = 600,
                     ready_after_execute: bool = True,
                     not_busy: bool = True):
        """
        wait max ready_before_execute seconds for pump to be ready
        move plunger to absolute position
        set post_confirmation to True to confirm plunger has moved to designated position
        USE WITH CARE
            set not_busy to False to not setting pump status to busy during moving
        Parameters
        ----------
        plunger_pos
        confirmation
        not_busy
        ready_before_execute
        ready_after_execute

        Returns
        -------

        """
        assert self.initialization_status, f'Pump {self.address} has not been initialized.'
        assert 0 <= plunger_pos <= self._max_step, "Plunger position out of range."
        if ready_before_execute > 0:
            self._ensure_ready(ready_before_execute)
        if not_busy:
            self._execute_pump(f'A{plunger_pos}')
        else:
            self._execute_pump(f'a{plunger_pos}')
        # print(f'Plunger moving to position {plunger_pos}.')
        if ready_after_execute:
            self._ensure_ready()
        if confirmation:
            if plunger_pos -5 <= self.report_plunger_absolute_position() <= plunger_pos + 5:
                self._update_plunger_absolute_position(plunger_pos)
                print(f'Pump {self.address} plunger moved to {self._plunger_absolute_position}.')
            else:
                raise IOError(f'Pump {self.address} move plunger failed!')
        else:
            self._update_plunger_absolute_position(self.report_plunger_absolute_position())

    @property
    def volume(self) -> float:
        """
        concrete method to report the current volume of the plunger
        Returns
        -------

        """
        plunger_pos = self.report_plunger_absolute_position()
        #volume = 1.0e3 * self.syringe_volume * (plunger_pos - self.home_pos) / self._max_step
        volume = 1.0e3 * self.syringe_volume * plunger_pos / self._max_step
        return volume

    @volume.setter
    def volume(self, volume: float, **kwargs):
        """
        concrete method to set the pistion_position
        convert the piston_position in mL (float) to plunger_pos in steps (int)
            based on syringe_volume in L (float) and max_step (int)
        move plunger to designated position
        Parameters
        ----------
        volume
        kwargs

        Returns
        -------

        """
        plunger_pos = int((1e-3 * volume / self.syringe_volume) * self._max_step)
        self.move_plunger(plunger_pos = plunger_pos, **kwargs)

    @property
    def port(self) -> int:
        """
        concrete method to report the current valve port number
        Returns
        -------
        current valve port number

        """
        return self.report_valve_number()

    @port.setter
    def port(self, port: int, **kwargs):
        """
        concrete method to set the valve number
        Parameters
        ----------
        port

        Returns
        -------

        """
        self.move_valve(valve_num = port, **kwargs)

    @property
    def top_speed_ml(self) -> float:
        """
        concrete method to report the top speed of plunger
        Returns
        -------

        """
        return self.report_top_speed() * 0.15 * self.syringe_volume

    @top_speed_ml.setter
    def top_speed_ml(self, top_speed: float):
        """
        concrete method to set the top speed
        Parameters
        ----------
        top_speed

        Returns
        -------

        """
        top_speed_int = int(top_speed / (0.15 * self.syringe_volume))
        self.set_top_speed(top_speed=top_speed_int)


# pump = TecanXCPump(com_port="COM6", address=0, syringe_volume=1e-3, num_valve_port=12)



from .jkem_pump import JKemPump
from .tecan_xc_pump import TecanXCPump
from .tecan_xlp_pump import TecanXLPPump
from .tecan_centris_pump import TecanCentrisPump
from .runze_pump import RunzePump
from .longer_peri import LongerPeristalticPump
from .continuous_dual_syringe import ContinuousDualSyringe
from .base_pump import SyringePump

__all__ = ["JKemPump", "TecanXCPump", "SyringePump", "TecanXLPPump", "TecanCentrisPump", "RunzePump", "LongerPeristalticPump", "ContinuousDualSyringe"]



from pathlib import Path
import time
from typing import Dict, Union, Tuple, Optional, List
import struct

from matterlab_serial_device import SerialDevice, open_close

from matterlab_pumps.base_pump import SyringePump


class JKemPump(SyringePump, SerialDevice):
    category="Pump"
    ui_fields  = ("com_port", "address", "syringe_volume", "home_pos")
    def __init__(self,
                 com_port: str,
                 address: Union[str, int],
                 syringe_volume: float,
                 num_valve_port: int,
                 ports: Optional[Dict[str, int]] = None,
                 init_valve: int = 5,
                 out_valve: int = 5,
                 connect_hardware: bool = True,
                 encoding: str = "utf-8",
                 baudrate: int = 38400,
                 timeout: float = 1.0,
                 bytesize: int = 8,
                 parity: str = "none",
                 stopbits: int = 1,
                 top_speed_ml: float = 0.1,
                 home_pos: int = -14000,
                 home_pos_port: int = -6000,
                 plunger_length: int = 200000,
                 top_speed: int = 384,
                 start_speed: int = 256,
                 stop_speed: int = 256,
                 speed_slope: int = 256,
                 wait_time: float = 3600.0,
                 step_resolution: int = 4,
                 share_port: bool = True,
                 **kwargs
                 ):
        """

        :param com_port: COM port of the pump connected to, example: "COM1", "/tty/USB0"
        :param address: address of pump, set at the back of the pump, char '0' - 'E' or int 0-9
        :param ports: dict containing nickname and valve port number mapping
        :param change_def_setting_path: if load a different default setting json file,
                                        default is 'default_settings_tecan.json'
        :param syringe_volume: syringe volume in unit of L (NOT mL!!!)
        :param init_valve:  valve port num to draw from when initializing, typically air/N2, pure solvent or waste
        :param out_valve:   valve port num to dispense to when initializing, typically waste
        :param connect_hardware: if initialize hardware up on connect
        """
        SerialDevice.__init__(self,
                              com_port=com_port,
                              encoding=encoding,
                              baudrate=baudrate,
                              timeout=timeout,
                              parity=parity,
                              bytesize=bytesize,
                              stopbits=stopbits,
                              share_port=share_port,
                              **kwargs,
                              )
        SyringePump.__init__(self,
                             address=address,
                             syringe_volume=syringe_volume,
                             num_valve_port=num_valve_port,
                             init_valve=init_valve,
                             out_valve=out_valve,
                             top_speed_ml=top_speed_ml,
                             home_pos=home_pos,
                             top_speed=top_speed,
                             start_speed=start_speed,
                             stop_speed=stop_speed,
                             speed_slope=speed_slope,
                             wait_time=wait_time,
                             ports=ports,
                             )
        self.home_pos_port = home_pos_port
        self.plunger_length = plunger_length
        self.step_resolution = step_resolution

        self.initialization_status = False
        if connect_hardware:
            self.connect()

    def connect(self):
        """
        hardware initialization function
        :return:
        """
        self._motor_reference = [False, False]
        self.initialization_status = False
        self._set_max_step()
        self._generate_valve_positions()
        self.initialize()

    @staticmethod
    def change_default_setting():
        setting_path = input("Please input new default setting json file path:\n")
        if Path(setting_path).exists() and Path(setting_path).is_file():
            return Path(setting_path)
        raise ValueError("Invalid setting path")

    @staticmethod
    def _update_address(address: Union[str, int]) -> int:
        """
        convert address of pump in str to int
        :param address:
        :return:
        """
        if isinstance(address, str):
            address = int(address)
        if 0 < address <= 16:
            return address
        else:
            raise ValueError("Address of pump must be 1-16!")

    @staticmethod
    def _checksum(command: Union[bytes, bytearray]) -> bytes:
        """
        calculate the checksum of a bytes or bytearray object and append the checksum to the end
        Parameters
        ----------
        command

        Returns
        -------

        """
        if isinstance(command, bytearray):
            command = bytes(command)
        sum = 0
        for command_byte in command:
            sum += command_byte
        checksum = command + struct.pack('B', sum & 0xFF)
        return checksum

    @open_close
    def _query_pump(self, instruction: int, command_type: int, motor: int, value: int) -> int:
        """
        query/execute the pump with a checksum_ed 9 bytes
        return None if query/execution failed
        return int value for querying
        for execution, return 0 (doesn't matter)

        Parameters
        ----------
        instruction
        command_type
        motor
        value

        Returns
        -------
        reading value in int
        """
        command = self._checksum(struct.pack('>BBBBi',
                                             self.address,
                                             instruction,
                                             command_type,
                                             motor,
                                             value))
        rtn = self.query(write_command = command, num_bytes = 9, return_bytes = True, read_delay = 0)
        if rtn[2] != 0x64:
            return None
        else:
            return struct.unpack('>i', rtn[4:8])[0]

    def _busy_report(self):
        """
        check if pump is ready for next command
            motor0 (syringe) and motor1(valve) both should have position_reached_flag == 1 (True)
        set self._ready to True if pump is ready
        set self._ready to False if pump is not ready
        Returns
        -------

        """
        motor0_status = self._query_pump(instruction = 6, command_type = 8, motor = 0, value = 0)
        motor1_status = self._query_pump(instruction=6, command_type=8, motor=1, value=0)
        if motor0_status and motor1_status:
            self._ready = True
        else:
            self._ready = False

    def _ensure_ready(self, wait_time: float = None) -> bool:
        """
        ensure the pump is ready
        Parameters
        ----------
        wait_time: max time to wait until pump is ready

        Returns: True if pump is ready
        -------

        """
        if not wait_time:
            wait_time = self.wait_time
        t0 = time.time()
        while (time.time() - t0) < wait_time:
            self._busy_report()
            if self._ready:
                return True
            else:
                time.sleep(0.05)
        raise IOError(f"Pump busy for {wait_time} seconds and not ready, exiting...")

    def _set_max_step(self):
        self._max_step = self.plunger_length + self.home_pos
    @property
    def max_step(self) -> int:
        return self._max_step

    def _set_home_position(self):
        """
        set the home position of the pump
        Parameters
        ----------

        Returns
        -------

        """
        #TODO: add two-step set home physically
        pass

    def report_plunger_absolute_position(self, max_try = 5) -> int:
        """
        report the current absolute position of the plunger
        Returns
        -------

        """
        for i in range (0, max_try):
            rtn = self._query_pump(instruction = 4, command_type = 1, motor = 0, value = 0)
            if isinstance(rtn, int):
                # print(f"Absolute plunger position is {rtn}.")
                return rtn
            time.sleep(1)
        raise IOError(f"Report plunger position of pump on {self.address} failed.")

    def _update_plunger_absolute_position(self, plunger_absolute_position: int):
        """
        update absolute plunger position internally
        Parameters
        ----------
        plunger_absolute_position

        Returns
        -------

        """
        self._plunger_absolute_position = plunger_absolute_position

    def _search_motor_reference(self, motor: int, max_wait_time = 120):
        """
        initialize the motor by searching for internal reference
        wait max_wait_time (default 120) s for finding the reference, otherwise raise error
        must follow by home
        Returns
        -------

        """
        rtn = self._query_pump(instruction = 13, command_type = 0, motor = motor, value = 0)
        if isinstance(rtn, int):
            for i in range(0, max_wait_time):
                if self._query_pump(instruction = 13, command_type = 2, motor = motor, value = 0) == 0:
                    self._motor_reference[motor] = True
                    return
                else:
                    time.sleep(1)
        raise IOError(f'Search reference of pump motor {motor} on {self.address} failed!')

    def initialize_plunger(self):
        """
        search for plunger reference (goto position 0)
        move plunger to home position
        wait until ready
        query current position, if in +/- 100 of setting,
        Returns
        -------

        """
        self._search_motor_reference(motor = 0)
        self._query_pump(instruction=4, command_type=0, motor=0, value=self.home_pos)
        self._ensure_ready(wait_time=120)
        plunger_pos = self.report_plunger_absolute_position()
        if plunger_pos in range(self.home_pos - 100, self.home_pos + 100):
            self._update_plunger_absolute_position(plunger_pos)
            print(f"Home plunger of pump on {self.address} succeed.")
            return
        raise IOError(f"Home plunger of pump on {self.address} failed.")

    def _generate_valve_positions(self):
        """
        create a list of motor positions of desired port
        self._valve_positions[0] port_1, etc.
        Returns
        -------

        """
        self._valve_positions = []
        for i in range(0, self.num_valve_port):
            pos = self.home_pos_port + int(i * 43928.0 / self.num_valve_port)
            self._valve_positions.append(pos)

    def report_valve_num(self, max_try = 5) -> int:
        """
        get the position of valve motor
        find match in +/- steps in the self_valve_positions
        Returns
        -------
        valve num in int, start from 1
        """
        for i in range(0, max_try):
            rtn = self._query_pump(instruction = 6, command_type = 1, motor = 1, value = 0)
            if isinstance(rtn, int):
                for j in range (0, self.num_valve_port):
                    if rtn in range(self._valve_positions[j] - 100, self._valve_positions[j] + 100):
                        print(f'Valve num of pump on {self.address} is {j+1}')
                        return j+1
            time.sleep(1)
        raise IOError(f'Report valve of pump on {self.address} failed.')

    def _update_valve_number(self, valve_number: int):
        """
        update the valve number internally
        Parameters
        ----------
        valve_number

        Returns
        -------

        """
        self._valve_number = valve_number

    def initialize_valve(self, init_valve: int = None):
        """

        Returns
        -------

        """
        if init_valve is None:
            init_valve = self.init_valve
        self._search_motor_reference(motor=1)
        self._query_pump(instruction=4, command_type=0, motor=1, value=self._valve_positions[init_valve - 1])
        self._ensure_ready(wait_time=30)
        valve_number = self.report_valve_num()
        if valve_number == init_valve:
            self._update_valve_number(valve_number)
            print(f"Initialize valve of pump on {self.address} succeed.")
            return
        raise IOError(f"Home valve of pump on {self.address} failed.")

    def _stop_motor(self, motor: int):
        """
        stop motor running
        Parameters
        ----------
        motor

        Returns
        -------

        """
        self._query_pump(instruction=3, command_type=0, motor=motor, value= 0)

    def stop_plunger(self):
        """
        stop the plunger moving
        Returns
        -------

        """
        self._stop_motor(motor = 0)

    def stop_valve(self):
        """
        stop the valve moving
        Returns
        -------

        """
        self._stop_motor(motor = 1)

    def _set_current(self, max_current = 100, standby_current = 20):
        """
        set the max current, do not change
        Parameters
        ----------
        max_current

        Returns
        -------

        """
        self._query_pump(instruction = 5, command_type = 6, motor = 0, value = max_current)
        self._query_pump(instruction=5, command_type=6, motor=1, value=max_current)
        self._query_pump(instruction=5, command_type=7, motor=0, value=standby_current)
        self._query_pump(instruction=5, command_type=7, motor=1, value=standby_current)

    def _set_interrupt_parameter(self):
        """
        set the interrupt parameter, do not change
        Returns
        -------

        """
        self._query_pump(instruction=9, command_type=27, motor=3, value=1)
        self._query_pump(instruction=9, command_type=29, motor=3, value=1)

    def _set_communication_parameter(self):
        """
        set the RS485 communication parameter, do not change
        Returns
        -------

        """
        self._query_pump(instruction=9, command_type=76, motor=0, value=255)

    def _set_left_switch_polarity(self):
        """
        set both motor left limit to active
        motor will stop if input is low
        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=13, motor=0, value=1)
        self._query_pump(instruction=5, command_type=13, motor=1, value=1)

    def _set_reference_search_mode(self):
        """
        set the reference search mode of both motor to search left switch only
        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=193, motor=0, value=1)
        self._query_pump(instruction=5, command_type=193, motor=1, value=1)

    def _set_reference_search_speed(self, speed: int = 128):
        """
        set the reference search speed of both motor,
        default search speed*2 = 256, switch speed 128
        Parameters
        ----------
        speed

        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=194, motor=0, value=speed*2)
        self._query_pump(instruction=5, command_type=194, motor=1, value=speed*2)
        self._query_pump(instruction=5, command_type=195, motor=0, value=speed)
        self._query_pump(instruction=5, command_type=195, motor=1, value=speed)

    def _set_boost_current(self, boost_current = 200):
        """
        set the boost current of plunger motor
        default 200
        Parameters
        ----------
        boost_current

        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=200, motor=0, value=boost_current)

    def _set_valve_max_positioning_speed(self, speed = 1024):
        """
        set the max positioning speed of valve motor
        default 1024
        Parameters
        ----------
        speed

        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=4, motor=1, value=speed)

    def _set_output(self):
        """
        set digital output
        not clear with trianic manual
        Returns
        -------

        """
        self._query_pump(instruction=14, command_type=0, motor=0, value=24)
        self._query_pump(instruction=14, command_type=0, motor=0, value=56)

    def report_top_speed(self, max_try = 5) -> int:
        """
        report the top speed of the plunger
        Returns
        -------

        """
        for i in range (0, max_try):
            rtn = self._query_pump(instruction=6, command_type=4, motor=0, value=0)
            if isinstance(rtn, int):
                print(f'Top speed of plunger on pump on {self.address} is {rtn}')
                return rtn
            time.sleep(1)
        raise IOError(f"Report top speed of plunger on pump on {self.address} failed.")

    def _update_top_speed(self, top_speed):
        """
        update the topspeed value internally
        Parameters
        ----------
        top_speed

        Returns
        -------

        """
        self._top_speed = top_speed

    def set_top_speed(self, top_speed: int):
        """
        set the speed of the plunger motor
        Parameters
        ----------
        top_speed

        Returns
        -------

        """
        assert 1<= top_speed <= 2047, "Top speed out of range, min 1, max 2047."
        self._query_pump(instruction=5, command_type=4, motor=0, value=top_speed)
        if self.report_top_speed() == top_speed:
            self._update_top_speed(top_speed)
        else:
            raise IOError("Set top speed failed!")

    def set_speed_slope(self, speed_slope: int):
        """
        set the acceleration parameters of both motors
        Parameters
        ----------
        speed_slope

        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=5, motor=0, value=speed_slope)
        self._query_pump(instruction=5, command_type=153, motor=0, value=6)
        self._query_pump(instruction=5, command_type=154, motor=0, value=3)
        self._query_pump(instruction=5, command_type=5, motor=1, value=speed_slope)
        self._query_pump(instruction=5, command_type=153, motor=1, value=6)
        self._query_pump(instruction=5, command_type=154, motor=1, value=3)

    def _set_step_resolution(self, step_resolution: int):
        """
        set the step resolution of plunger motor
        default 4
        Parameters
        ----------
        step_resolution

        Returns
        -------

        """
        self._query_pump(instruction=5, command_type=160, motor=0, value=0)
        self._query_pump(instruction=5, command_type=140, motor=0, value=step_resolution)
        if step_resolution == 4:
            self._query_pump(instruction=5, command_type=160, motor=0, value=1)

    def _initialize_setting(self):
        """
        test communication with motor 0 (plunger motor)
        initialize the settings of the pump
        Returns
        -------

        """
        if self._query_pump(instruction = 6, command_type = 1, motor = 0, value = 0) is None:
            raise IOError(f"Initial communication with pump on {self.address} failed.")
        print(f"Start initializing pump {self.address}")
        self.stop_plunger()
        self.stop_valve()
        self._set_current()
        self._set_interrupt_parameter()
        self._set_communication_parameter()
        self._set_left_switch_polarity()
        self._set_reference_search_mode()
        self._set_reference_search_speed()
        self._set_boost_current()
        self._set_valve_max_positioning_speed()
        self._set_output()
        self.set_top_speed(self._top_speed_def)
        self.set_speed_slope(self._speed_slope_def)
        self._set_step_resolution(self.step_resolution)

    def initialize(self, init_valve: int = None):
        """
        initialize the pump by
            1. initialize the settings
            2. initialize the valve to init port
            3. initialize the plunger to home position
        set initialize status to True
        Returns
        -------

        """
        self._initialize_setting()
        self.initialize_valve(init_valve = init_valve)
        self.initialize_plunger()
        self.initialization_status = True
        print(f"Pump on {self.address} has been initialized")

    def move_plunger(self, plunger_pos: int,
                     confirmation: bool = True,
                     ready_before_execute: int = 600,
                     ready_after_execute: bool = True):
        assert (self.home_pos - 50) <= plunger_pos <= (self._max_step + 50), "Plunger position out of range!"
        if ready_before_execute > 0:
            self._ensure_ready(ready_before_execute)
        self._query_pump(instruction=4, command_type=0, motor=0, value= plunger_pos)
        if ready_after_execute:
            self._ensure_ready()
        if confirmation:
            if self.report_plunger_absolute_position() in range (plunger_pos -50, plunger_pos + 50):
                self._update_plunger_absolute_position(plunger_pos)
                print(f'Pump {self.address} plunger moved to {self._plunger_absolute_position}.')
            else:
                raise IOError(f'Pump {self.address} move plunger failed!')
        else:
            self._update_plunger_absolute_position(self.report_plunger_absolute_position())

    def move_valve(self, valve_num: int,
                   confirmation: bool = True,
                   ready_before_execute: int = 600):
        assert 1 <= valve_num <= 8, "Valve number out of range, min 1, max 8!"
        if ready_before_execute > 0:
            self._ensure_ready(ready_before_execute)
        self._query_pump(instruction=4, command_type=0, motor=1, value=self._valve_positions[valve_num - 1])
        self._ensure_ready(wait_time=3)
        if confirmation:
            if self.report_valve_num() == valve_num:
                self._update_valve_number(valve_num)
                print(f"Pump {self.address} valve moved to {self._valve_number}.")
            else:
                raise IOError(f"Pump {self.address} move valve failed.")
        else:
            self._update_valve_number(self.report_valve_num())

    @property
    def port(self) -> int:
        """
        concrete method to report the current valve port number
        Returns
        -------
        current valve port number

        """
        return self.report_valve_num()

    @port.setter
    def port(self, port: int, **kwargs) -> None:
        """
        concrete method to set the valve number
        Parameters
        ----------
        port
        kwargs

        Returns
        -------

        """
        self.move_valve(valve_num=port, **kwargs)

    @property
    def volume(self) -> float:
        """
        concrete method to report the current volume of the plunger
        Returns
        -------

        """
        plunger_pos = self.report_plunger_absolute_position()
        volume = 1.0e3 * self.syringe_volume * (
                    plunger_pos - self.home_pos) / self.plunger_length
        return volume

    @volume.setter
    def volume(self, volume: float, **kwargs):
        """
        concrete method to set the pistion_position
        convert the piston_position in mL (float) to plunger_pos in steps (int)
            based on syringe_volume in L (float) and max_step (int)
        move plunger to designated position
        Parameters
        ----------
        volume

        Returns
        -------

        """
        plunger_pos = int((1e-3 * volume / self.syringe_volume) * self.plunger_length) + self.home_pos
        self.move_plunger(plunger_pos = plunger_pos, **kwargs)

    @property
    def top_speed_ml(self)-> float:
        """
        concrete method to get the speed of plunger based on mL/s
        Returns
        -------

        """
        speed_step = self.report_top_speed()
        speed = 0.13 * self.syringe_volume * speed_step
        return speed
 
    @top_speed_ml.setter
    def top_speed_ml(self, speed: float):
        """
        concrete method to set the speed of plunger based on mL/s
        Parameters
        ----------
        speed

        Returns
        -------

        """
        speed_step = int(speed / (0.13 * self.syringe_volume))
        if speed_step > 2047:
            speed_step = 2047
        self.set_top_speed(top_speed = speed_step)

    def set_default_speeds(self):
        pass